import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    This notebook contains outputs of many different types: text, HTML, plots and errors.
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Text outputs

    Using `print`, `sys.stdout` and `sys.stderr`
    """)
    return


@app.cell
def _():
    import sys
    print('using print')
    sys.stdout.write('using sys.stdout.write')
    sys.stderr.write('using sys.stderr.write')
    return


@app.cell
def _():
    import logging
    logging.debug('Debug')
    logging.info('Info')
    logging.warning('Warning')
    logging.error('Error')
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # HTML outputs

    Using `pandas`. Here we find two representations: both text and HTML.
    """)
    return


@app.cell
def _():
    import pandas as pd
    pd.DataFrame([4])
    return (pd,)


@app.cell
def _(pd):
    from IPython.display import display
    display(pd.DataFrame([5]))
    display(pd.DataFrame([6]))
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Images
    """)
    return


@app.cell
def _():
    # '%matplotlib inline' command supported automatically in marimo
    return


@app.cell
def _():
    # First plot
    from matplotlib import pyplot as plt
    import numpy as np
    w, h = 3, 3
    data = np.zeros((h, w, 3), dtype=np.uint8)
    data[0,:] = [0,255,0]
    data[1,:] = [0,0,255]
    data[2,:] = [0,255,0]
    data[1:3,1:3] = [255, 0, 0]
    plt.imshow(data)
    plt.axis('off')
    plt.show()
    # Second plot
    data[1:3,1:3] = [255, 255, 0]
    plt.imshow(data)
    plt.axis('off')
    plt.show()
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Errors
    """)
    return


@app.cell
def _(undefined_variable):
    undefined_variable
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    A markdown cell
    """)
    return


@app.cell
def _():
    1+1
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Markdown cell two
    """)
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()



import os
from pathlib import Path
from contextlib import contextmanager

import pytest

from jupytext.config import (
    find_jupytext_configuration_file,
    load_jupytext_configuration_file,
    notebook_formats,
)
from jupytext.jupytext import load_jupytext_config, read


@contextmanager
def change_dir(path):
    """Context manager to temporarily change directory."""
    old_cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(old_cwd)


@pytest.fixture
def temp_folder_tree(tmp_path):
    """
    Fixture that creates a temporary folder tree.
    tmp_path is a pytest built-in fixture that auto-cleans up.
    """
    # Create folder structure
    (tmp_path / "subdir").mkdir()

    # Create some files
    (tmp_path / "jupytext.toml").write_text("")
    (tmp_path / "subdir" / "notebook.ipynb").write_text("")

    return tmp_path


def test_issue_1440(temp_folder_tree):
    """Test that jupytext finds the config file even in parent of cwd"""
    notebook = Path("notebook.ipynb")
    notebook_str = str(notebook)
    expected_toml = str(temp_folder_tree / "jupytext.toml")
    with change_dir(temp_folder_tree / "subdir"):
        assert notebook.exists()
        config_file_from_path = find_jupytext_configuration_file(notebook)
        assert config_file_from_path == expected_toml
        config_file_from_str = find_jupytext_configuration_file(notebook_str)
        assert config_file_from_str == expected_toml


def test_find_jupytext_configuration_file(tmpdir):
    nested = tmpdir.mkdir("nested")

    # Start with no configuration
    assert find_jupytext_configuration_file(str(nested)) is None

    # Configuration file in the parent directory
    root_config = tmpdir.join("jupytext.yml")
    root_config.write("\n")
    assert os.path.samefile(find_jupytext_configuration_file(str(tmpdir)), str(root_config))
    assert os.path.samefile(find_jupytext_configuration_file(str(nested)), str(root_config))

    # Local pyproject file
    pyproject_config = nested.join("pyproject.toml")
    pyproject_config.write("[tool.jupytext]\n")
    assert os.path.samefile(find_jupytext_configuration_file(str(tmpdir)), str(root_config))
    assert os.path.samefile(find_jupytext_configuration_file(str(nested)), str(pyproject_config))

    # Local configuration file
    local_config = nested.join(".jupytext")
    local_config.write("\n")
    assert os.path.samefile(find_jupytext_configuration_file(str(tmpdir)), str(root_config))
    assert os.path.samefile(find_jupytext_configuration_file(str(nested)), str(local_config))


def test_jupytext_py_is_not_a_configuration_file(tmpdir):
    jupytext_py = tmpdir.join("jupytext.py")
    jupytext_py.write("# Not a config file!")

    assert find_jupytext_configuration_file(str(tmpdir)) is None

    dot_jupytext_py = tmpdir.join(".jupytext.py")
    dot_jupytext_py.write("# This is a config file!")

    assert find_jupytext_configuration_file(str(tmpdir)) == str(dot_jupytext_py)


@pytest.mark.parametrize(
    "config_file",
    [
        "pyproject.toml",
        "jupytext",
        "jupytext.toml",
        "jupytext.yml",
        "jupytext.json",
        "jupytext.py",
    ],
)
def test_load_jupytext_configuration_file(tmpdir, config_file):
    full_config_path = tmpdir.join(config_file)

    if config_file == "pyproject.toml":
        full_config_path.write(
            """[tool.jupytext]
formats = "ipynb,py:percent"
notebook_metadata_filter = "all"
cell_metadata_filter = "all"
"""
        )
    elif config_file.endswith(("jupytext", ".toml")):
        full_config_path.write(
            """formats = "ipynb,py:percent"
notebook_metadata_filter = "all"
cell_metadata_filter = "all"
"""
        )
    elif config_file.endswith(".yml"):
        full_config_path.write(
            """formats: ipynb,py:percent
notebook_metadata_filter: all
cell_metadata_filter: all
"""
        )
    elif config_file.endswith(".json"):
        full_config_path.write(
            """{"formats": "ipynb,py:percent",
"notebook_metadata_filter": "all",
"cell_metadata_filter": "all"
}
"""
        )
    elif config_file.endswith(".py"):
        full_config_path.write(
            """c.formats = "ipynb,py:percent"
c.notebook_metadata_filter = "all"
c.cell_metadata_filter = "all"
"""
        )

    config = load_jupytext_configuration_file(str(full_config_path))
    assert config.formats == ["ipynb,py:percent"]
    assert config.notebook_metadata_filter == "all"
    assert config.cell_metadata_filter == "all"


@pytest.mark.parametrize(
    "content_toml,formats_short_form",
    [
        (
            """# always pair ipynb notebooks to py:percent files
formats = "ipynb,py:percent"
""",
            "ipynb,py:percent",
        ),
        (
            """# always pair ipynb notebooks to py:percent files
formats = "ipynb,py:percent"
""",
            "ipynb,py:percent",
        ),
        (
            """# Pair notebooks in subfolders of 'notebooks' to scripts in subfolders of 'scripts'
[formats]
"notebooks/" = "ipynb"
"scripts/" = "py:percent"
""",
            "notebooks///ipynb,scripts///py:percent",
        ),
        (
            """# Pair notebooks in subfolders of 'notebooks' to scripts in subfolders of 'scripts'
[formats]
"notebooks" = "ipynb"
"scripts" = "py:percent"
""",
            "notebooks///ipynb,scripts///py:percent",
        ),
        (
            """# Pair local notebooks to scripts in 'notebooks_py' and md files in 'notebooks_md'
[formats]
"" = "ipynb"
"notebooks_py" = "py:percent"
"notebooks_md" = "md:myst"
""",
            "ipynb,notebooks_py///py:percent,notebooks_md///md:myst",
        ),
    ],
)
def test_jupytext_formats(tmpdir, content_toml, formats_short_form):
    jupytext_toml = tmpdir.join("jupytext.toml")
    jupytext_toml.write(content_toml)

    config = load_jupytext_configuration_file(str(jupytext_toml))
    assert config.formats == [formats_short_form]


def test_deprecated_formats_cause_warning(tmpdir, content_toml="default_jupytext_formats = 'ipynb,md'"):
    jupytext_toml = tmpdir.join("jupytext.toml")
    jupytext_toml.write(content_toml)

    config = load_jupytext_configuration_file(str(jupytext_toml))
    with pytest.warns(FutureWarning, match="use 'formats'"):
        assert config.default_formats(str(tmpdir.join("test.md"))) == "ipynb,md"


@pytest.mark.parametrize(
    "option_name",
    ["notebook_metadata_filter", "cell_metadata_filter", "cell_markers"],
)
def test_deprecated_options_cause_warning(tmpdir, option_name):
    jupytext_toml = tmpdir.join("jupytext.toml")
    jupytext_toml.write(f"default_{option_name} = 'value'")
    config = load_jupytext_configuration_file(str(jupytext_toml))
    fmt = {}
    with pytest.warns(FutureWarning, match=f"use '{option_name}'"):
        config.set_default_format_options(fmt)
        assert fmt[option_name] == "value"


def test_simple_py_file_is_not_paired(tmp_path):
    py_file = tmp_path / "simple.py"
    py_file.write_text('print("Hello, world!")')

    config_file = tmp_path / "jupytext.toml"
    config_file.write_text('formats = "ipynb,py:percent"')

    notebook = read(str(py_file))
    config_file = load_jupytext_config(str(config_file))

    formats = notebook_formats(notebook, config_file, str(py_file))
    assert formats == [{"extension": ".py", "format_name": "light"}], formats


def test_pairing_groups(tmp_path):
    """Test list-based formats for subset-specific pairing"""
    jupytext_toml = tmp_path / "jupytext.toml"
    jupytext_toml.write_text("""
[[formats]]
"notebooks/tutorials/" = "ipynb"
"docs/tutorials/" = "md"
"scripts/tutorials/" = "py:percent"

[[formats]]
"notebooks/" = "ipynb"
"scripts/" = "py:percent"
""")

    config = load_jupytext_configuration_file(str(jupytext_toml))

    # Formats should be a list of str
    assert isinstance(config.formats, list)
    assert len(config.formats) == 2
    assert all(isinstance(f, str) for f in config.formats)

    # Test that default_formats returns the correct formats based on path
    # Tutorial notebook should match first format (tutorials)
    tutorial_notebook = str(tmp_path / "notebooks" / "tutorials" / "getting_started.ipynb")
    assert (
        config.default_formats(tutorial_notebook)
        == "notebooks/tutorials///ipynb,docs/tutorials///md,scripts/tutorials///py:percent"
    )

    # Regular notebook should match second format (main)
    regular_notebook = str(tmp_path / "notebooks" / "hello.ipynb")
    assert config.default_formats(regular_notebook) == "notebooks///ipynb,scripts///py:percent"


def test_pairing_groups_multiple_groups(tmp_path):
    """Test multiple format sets with list-based formats"""
    jupytext_toml = tmp_path / "jupytext.toml"
    jupytext_toml.write_text("""
[[formats]]
"notebooks/tutorials/" = "ipynb"
"docs/tutorials/" = "md"

[[formats]]
"notebooks/examples/" = "ipynb"
"docs/examples/" = "md:myst"

[[formats]]
"notebooks/" = "ipynb"
"scripts/" = "py:percent"
""")

    config = load_jupytext_configuration_file(str(jupytext_toml))

    # Check formats is a list of dicts
    assert isinstance(config.formats, list)
    assert len(config.formats) == 3

    # Test that the correct format is selected based on path (first match wins)
    tutorial_notebook = str(tmp_path / "notebooks" / "tutorials" / "intro.ipynb")
    assert config.default_formats(tutorial_notebook) == "notebooks/tutorials///ipynb,docs/tutorials///md"

    example_notebook = str(tmp_path / "notebooks" / "examples" / "demo.ipynb")
    assert config.default_formats(example_notebook) == "notebooks/examples///ipynb,docs/examples///md:myst"

    # Regular notebook should use the last (default) format
    regular_notebook = str(tmp_path / "notebooks" / "regular.ipynb")
    assert config.default_formats(regular_notebook) == "notebooks///ipynb,scripts///py:percent"


def test_formats_list_without_default(tmp_path):
    """Test list-based formats without a default catchall"""
    jupytext_toml = tmp_path / "jupytext.toml"
    jupytext_toml.write_text("""
[[formats]]
"notebooks/tutorials/" = "ipynb"
"docs/tutorials/" = "md"
""")

    config = load_jupytext_configuration_file(str(jupytext_toml))

    # Formats should be a list
    assert isinstance(config.formats, list)
    assert len(config.formats) == 1

    # Matching notebook should work
    tutorial_notebook = str(tmp_path / "notebooks" / "tutorials" / "getting_started.ipynb")
    assert config.default_formats(tutorial_notebook) == "notebooks/tutorials///ipynb,docs/tutorials///md"

    # Non-matching notebook should return None
    other_notebook = str(tmp_path / "notebooks" / "other.ipynb")
    assert config.default_formats(other_notebook) is None


def test_formats_list_yaml(tmp_path):
    """Test list-based formats with YAML config"""
    jupytext_yml = tmp_path / "jupytext.yml"
    jupytext_yml.write_text("""
formats:
  - notebooks/tutorials/: ipynb
    docs/tutorials/: md
  - notebooks/: ipynb
    scripts/: py:percent
""")

    config = load_jupytext_configuration_file(str(jupytext_yml))

    # Formats should be a list
    assert isinstance(config.formats, list)
    assert len(config.formats) == 2


def test_formats_list_json(tmp_path):
    """Test list-based formats with JSON config"""
    jupytext_json = tmp_path / "jupytext.json"
    jupytext_json.write_text("""{
  "formats": [
    {
      "notebooks/tutorials/": "ipynb",
      "docs/tutorials/": "md"
    },
    {
      "notebooks/": "ipynb",
      "scripts/": "py:percent"
    }
  ]
}
""")

    config = load_jupytext_configuration_file(str(jupytext_json))

    # Formats should be a list
    assert isinstance(config.formats, list)
    assert len(config.formats) == 2


def test_formats_semicolon_separated(tmp_path):
    """Test semicolon-separated format strings"""
    jupytext_toml = tmp_path / "jupytext.toml"
    jupytext_toml.write_text("""
formats = "notebooks///ipynb,scripts///py:percent;ipynb,py:percent"
""")

    config = load_jupytext_configuration_file(str(jupytext_toml))

    # Formats should be split into a list
    assert isinstance(config.formats, list)
    assert len(config.formats) == 2
    assert config.formats[0] == "notebooks///ipynb,scripts///py:percent"
    assert config.formats[1] == "ipynb,py:percent"

    # Test that the correct format is selected based on path
    notebook_in_notebooks = str(tmp_path / "notebooks" / "test.ipynb")
    assert config.default_formats(notebook_in_notebooks) == "notebooks///ipynb,scripts///py:percent"

    notebook_elsewhere = str(tmp_path / "test.ipynb")
    assert config.default_formats(notebook_elsewhere) == "ipynb,py:percent"


def test_formats_toml_list_of_strings(tmp_path):
    """Test TOML list with string format specifications"""
    jupytext_toml = tmp_path / "jupytext.toml"
    jupytext_toml.write_text("""
formats = [
    "notebooks///ipynb,scripts///py:percent",
    "ipynb,py:percent"
]
""")

    config = load_jupytext_configuration_file(str(jupytext_toml))

    # Formats should be a list
    assert isinstance(config.formats, list)
    assert len(config.formats) == 2
    assert config.formats[0] == "notebooks///ipynb,scripts///py:percent"
    assert config.formats[1] == "ipynb,py:percent"

    # Test that the correct format is selected based on path
    notebook_in_notebooks = str(tmp_path / "notebooks" / "test.ipynb")
    assert config.default_formats(notebook_in_notebooks) == "notebooks///ipynb,scripts///py:percent"

    notebook_elsewhere = str(tmp_path / "test.ipynb")
    assert config.default_formats(notebook_elsewhere) == "ipynb,py:percent"



from jupytext.marimo import marimo_py_to_notebook, notebook_to_marimo_py, marimo_version
import pytest
from nbformat.v4.nbbase import new_notebook, new_code_cell
from jupytext.compare import compare, compare_notebooks
from jupytext.formats import guess_format


@pytest.fixture
def py_marimo() -> str:
    return f"""import marimo

__generated_with = "{marimo_version()}"
app = marimo.App()


@app.cell
def _():
    x = 1
    return


if __name__ == "__main__":
    app.run()"""


@pytest.fixture
def notebook():
    return new_notebook(cells=[new_code_cell("x = 1")])


def test_guess_format(py_marimo):
    assert guess_format(py_marimo, ".py") == ("marimo", {})


@pytest.mark.requires_marimo
def test_notebook_to_py_marimo(py_marimo, notebook):
    actual = notebook_to_marimo_py(notebook)
    compare(actual, py_marimo)


@pytest.mark.requires_marimo
def test_marimo_py_to_notebook(py_marimo, notebook):
    actual = marimo_py_to_notebook(py_marimo)
    compare_notebooks(actual, notebook)



"""Jupyter notebook to Quarto Markdown and back, using 'quarto convert'"""

import json
import os
import shutil
import subprocess
import sys
import tempfile

# Copy nbformat reads and writes to avoid them being patched in the contents manager!!
from nbformat import reads as ipynb_reads
from nbformat import writes as ipynb_writes
from packaging.version import parse
import yaml

QUARTO_MIN_VERSION = "0.2.134"


class QuartoError(OSError):
    """An error related to Quarto"""


def quarto(args, filein=None):
    """Execute quarto with the given arguments"""
    executable = shutil.which("quarto")
    if not executable and sys.platform.startswith("win"):
        # On Windows, try quarto.cmd, see #1406
        executable = shutil.which("quarto.cmd")
    if not executable:
        raise QuartoError("Could not find 'quarto' executable")

    cmd = [executable] + args.split()

    if filein:
        cmd.append(filein)

    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode:
        raise QuartoError(f"{' '.join(cmd)} exited with return code {proc.returncode}\n{err.decode('utf-8')}")

    return out.decode("utf-8")


def is_quarto_available(min_version=QUARTO_MIN_VERSION):
    """Is Quarto available?"""
    try:
        raise_if_quarto_is_not_available(min_version=min_version)
        return True
    except QuartoError:
        return False


def raise_if_quarto_is_not_available(min_version=QUARTO_MIN_VERSION):
    """Raise with an informative error message if quarto is not available"""
    version = quarto_version()
    if version == "N/A":
        raise QuartoError(f"The Quarto Markdown format requires 'quarto>={min_version}', but quarto was not found")

    if parse(version) < parse(min_version):
        raise QuartoError(
            f"The Quarto Markdown format requires 'quarto>={min_version}', but quarto version {version} was found"
        )

    return version


def quarto_version():
    """Quarto's version number"""
    try:
        return quarto("--version").strip()
    except OSError:
        return "N/A"


def qmd_to_notebook(text):
    """Convert a Quarto Markdown notebook to a Jupyter notebook"""
    raise_if_quarto_is_not_available()
    tmp_qmd_file = tempfile.NamedTemporaryFile(delete=False, suffix=".qmd")
    tmp_qmd_file.write(text.encode("utf-8"))
    tmp_qmd_file.close()

    quarto("convert --log-level warning", tmp_qmd_file.name)

    tmp_ipynb_file_name = tmp_qmd_file.name[:-4] + ".ipynb"

    with open(tmp_ipynb_file_name, encoding="utf-8") as ipynb_file:
        notebook = ipynb_reads(ipynb_file.read(), as_version=4)

    os.unlink(tmp_qmd_file.name)
    os.unlink(tmp_ipynb_file_name)

    # Recent version of Quarto duplicate the notebook metadata in a YAML header
    # If we find such a header, we remove it if it is identical to the notebook metadata
    if not notebook.cells or notebook.cells[0].cell_type != "markdown":
        return notebook

    cell_source = notebook.cells[0].source.strip()
    if cell_source.startswith("---") and cell_source.endswith("---"):
        try:
            metadata_from_top_cell = yaml.safe_load(cell_source[3:-3])
        except yaml.YAMLError:
            pass
        else:
            if "jupyter" not in metadata_from_top_cell:
                return notebook
            if metadata_from_top_cell == {"jupyter": "python3"}:
                # Quarto sometimes writes "jupyter: python3" instead of the full metadata
                # We can safely remove this cell
                del notebook.cells[0]
                return notebook

            jupyter_metadata = json.dumps(metadata_from_top_cell["jupyter"], sort_keys=True)
            notebook_metadata = json.dumps(notebook.metadata, sort_keys=True)
            if jupyter_metadata == notebook_metadata:
                if set(metadata_from_top_cell.keys()) == {"jupyter"}:
                    # The metadata are identical, we can remove the duplicated cell
                    del notebook.cells[0]
                    return notebook
                else:
                    # The jupyter metadata are identical, but there are other keys in the YAML header
                    # We keep the cell, but remove the jupyter key
                    lines = cell_source.splitlines()
                    non_jupyter_lines = []
                    in_jupyter_section = False
                    for line in lines:
                        if not (line.startswith("  ") or line.startswith("\t")):
                            in_jupyter_section = False
                        if line.rstrip() == "jupyter:":
                            in_jupyter_section = True
                        if not in_jupyter_section:
                            non_jupyter_lines.append(line)
                    notebook.cells[0].source = "\n".join(non_jupyter_lines)
            else:
                # The metadata are different, we keep the cell but issue a warning
                print(
                    "Warning: the Quarto Markdown file contains a YAML header that does not match the notebook metadata:\n"
                    f"Header metadata:\n{metadata_from_top_cell}\n"
                    f"Notebook metadata:\n{notebook_metadata}",
                    file=sys.stderr,
                )

    return notebook


def notebook_to_qmd(notebook):
    """Convert a Jupyter notebook to its Quarto Markdown representation"""
    raise_if_quarto_is_not_available()
    tmp_ipynb_file = tempfile.NamedTemporaryFile(delete=False, suffix=".ipynb")
    tmp_ipynb_file.write(ipynb_writes(notebook).encode("utf-8"))
    tmp_ipynb_file.close()

    quarto("convert --log-level warning", tmp_ipynb_file.name)

    tmp_qmd_file_name = tmp_ipynb_file.name[:-6] + ".qmd"

    with open(tmp_qmd_file_name, encoding="utf-8") as qmd_file:
        text = qmd_file.read()

    os.unlink(tmp_ipynb_file.name)
    os.unlink(tmp_qmd_file_name)

    return "\n".join(text.splitlines())



# ---
#
# This is a complex paragraph
# that is split over multiple lines.
#
# It also includes blank lines.
#
#
# jupyter:
#   jupytext:
#     cell_markers: '{{{,}}}'
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

print("Hello, World!")



"""Compare two Jupyter notebooks"""

import difflib
import json
import os
import re

from jupytext.marimo import marimo_version
from jupytext.paired_paths import full_path

from .cell_metadata import _IGNORE_CELL_METADATA
from .combine import combine_inputs_with_outputs
from .formats import check_auto_ext, long_form_one_format
from .header import _DEFAULT_NOTEBOOK_METADATA
from .jupytext import read, reads, write, writes
from .metadata_filter import filter_metadata

_BLANK_LINE = re.compile(r"^\s*$")


def _multilines(obj):
    try:
        lines = obj.splitlines()
        return lines + [""] if obj.endswith("\n") else lines
    except AttributeError:
        # Remove the final blank space on Python 2.7
        # return json.dumps(obj, indent=True, sort_keys=True).splitlines()
        return [line.rstrip() for line in json.dumps(obj, indent=True, sort_keys=True).splitlines()]


def compare(actual, expected, actual_name="actual", expected_name="expected", return_diff=False):
    """Compare two strings, lists or dict-like objects"""
    if actual != expected:
        diff = difflib.unified_diff(
            _multilines(expected),
            _multilines(actual),
            expected_name,
            actual_name,
            lineterm="",
        )
        if expected_name == "" and actual_name == "":
            diff = list(diff)[2:]
        diff = "\n".join(diff)
        if return_diff:
            return diff
        raise AssertionError("\n" + diff)
    return "" if return_diff else None


def filtered_cell(cell, preserve_outputs, cell_metadata_filter):
    """Cell type, metadata and source from given cell"""
    filtered = {
        "cell_type": cell.cell_type,
        "source": cell.source,
        "metadata": filter_metadata(cell.metadata, cell_metadata_filter, _IGNORE_CELL_METADATA),
    }

    if preserve_outputs:
        for key in ["execution_count", "outputs"]:
            if key in cell:
                filtered[key] = cell[key]

    return filtered


def filtered_notebook_metadata(notebook, ignore_kernelspec=False):
    """Notebook metadata, filtered for metadata added by Jupytext itself"""
    metadata = filter_metadata(
        notebook.metadata,
        notebook.metadata.get("jupytext", {}).get("notebook_metadata_filter"),
        _DEFAULT_NOTEBOOK_METADATA,
    )

    # Quarto round-trips may change the kernelspec
    if ignore_kernelspec:
        metadata.pop("kernelspec", None)

    if "jupytext" in metadata:
        del metadata["jupytext"]
    return metadata


class NotebookDifference(Exception):
    """Report notebook differences"""


def same_content(ref_source, test_source, allow_removed_final_blank_line):
    """Is the content of two cells the same, except for an optional final blank line?"""
    if ref_source == test_source:
        return True

    if not allow_removed_final_blank_line:
        return False

    # Is ref identical to test, plus one blank line?
    ref_source = ref_source.splitlines()
    test_source = test_source.splitlines()

    if not ref_source:
        return False

    if ref_source[:-1] != test_source:
        return False

    return _BLANK_LINE.match(ref_source[-1])


def compare_notebooks(
    notebook_actual,
    notebook_expected,
    fmt=None,
    allow_expected_differences=True,
    raise_on_first_difference=True,
    compare_outputs=False,
    compare_ids=None,
):
    """Compare the two notebooks, and raise with a meaningful message
    that explains the differences, if any"""
    fmt = long_form_one_format(fmt)
    format_name = fmt.get("format_name")

    if format_name == "sphinx" and notebook_actual.cells and notebook_actual.cells[0].source == "%matplotlib inline":
        notebook_actual.cells = notebook_actual.cells[1:]

    if compare_ids is None:
        compare_ids = compare_outputs

    modified_cells, modified_cell_metadata = compare_cells(
        notebook_actual.cells,
        notebook_expected.cells,
        raise_on_first_difference,
        compare_outputs=compare_outputs,
        compare_ids=compare_ids,
        cell_metadata_filter="-all"
        if format_name == "marimo"
        else notebook_actual.get("jupytext", {}).get("cell_metadata_filter"),
        allow_missing_code_cell_metadata=(allow_expected_differences and format_name in ["sphinx", "marimo"]),
        allow_missing_markdown_cell_metadata=(allow_expected_differences and format_name in ["sphinx", "spin", "marimo"]),
        allow_filtered_cell_metadata=allow_expected_differences,
        allow_removed_final_blank_line=allow_expected_differences,
    )

    # Compare notebook metadata
    modified_metadata = False
    if fmt.get("format_name") != "marimo":
        try:
            ignore_kernelspec = fmt.get("extension") == ".qmd" and allow_expected_differences
            compare(
                filtered_notebook_metadata(notebook_actual, ignore_kernelspec),
                filtered_notebook_metadata(notebook_expected, ignore_kernelspec),
            )
        except AssertionError as error:
            if raise_on_first_difference:
                raise NotebookDifference(f"Notebook metadata differ: {str(error)}")
            modified_metadata = True

    error = []
    if modified_cells:
        error.append(
            "Cells {} differ ({}/{})".format(
                ",".join([str(i) for i in modified_cells]),
                len(modified_cells),
                len(notebook_expected.cells),
            )
        )
    if modified_cell_metadata:
        error.append("Cell metadata '{}' differ".format("', '".join([str(i) for i in modified_cell_metadata])))
    if modified_metadata:
        error.append("Notebook metadata differ")

    if error:
        raise NotebookDifference(" | ".join(error))


def compare_cells(
    actual_cells,
    expected_cells,
    raise_on_first_difference=True,
    compare_outputs=True,
    compare_ids=True,
    cell_metadata_filter=None,
    allow_missing_code_cell_metadata=False,
    allow_missing_markdown_cell_metadata=False,
    allow_filtered_cell_metadata=False,
    allow_removed_final_blank_line=False,
):
    """Compare two collection of notebook cells"""
    test_cell_iter = iter(actual_cells)
    modified_cells = set()
    modified_cell_metadata = set()

    for i, ref_cell in enumerate(expected_cells, 1):
        try:
            test_cell = next(test_cell_iter)
        except StopIteration:
            if raise_on_first_difference:
                raise NotebookDifference(f"No cell corresponding to {ref_cell.cell_type} cell #{i}:\n{ref_cell.source}")
            modified_cells.update(range(i, len(expected_cells) + 1))
            break

        ref_lines = [line for line in ref_cell.source.splitlines() if not _BLANK_LINE.match(line)]
        test_lines = []

        # 1. test cell type
        if ref_cell.cell_type != test_cell.cell_type:
            if raise_on_first_difference:
                raise NotebookDifference(
                    f"When comparing cell #{i}: "
                    f"expecting a {ref_cell.cell_type} cell, but got a {test_cell.cell_type} cell.\n"
                    f"Expected content:\n{ref_cell.source}\nActual content:\n{test_cell.source}"
                )
            modified_cells.add(i)

        # Compare cell ids (introduced in nbformat 5.1.0)
        if compare_ids and test_cell.get("id") != ref_cell.get("id"):
            if raise_on_first_difference:
                raise NotebookDifference(
                    f"Cell ids differ on {test_cell['cell_type']} cell #{i}: '{test_cell.get('id')}' != '{ref_cell.get('id')}'"
                )
            modified_cells.add(i)

        # 2. test cell metadata
        if (ref_cell.cell_type == "code" and not allow_missing_code_cell_metadata) or (
            ref_cell.cell_type != "code" and not allow_missing_markdown_cell_metadata
        ):
            ref_metadata = ref_cell.metadata
            test_metadata = test_cell.metadata
            if allow_filtered_cell_metadata:
                ref_metadata = {key: ref_metadata[key] for key in ref_metadata if key not in _IGNORE_CELL_METADATA}
                test_metadata = {key: test_metadata[key] for key in test_metadata if key not in _IGNORE_CELL_METADATA}

            if ref_metadata != test_metadata:
                if raise_on_first_difference:
                    try:
                        compare(test_metadata, ref_metadata)
                    except AssertionError as error:
                        raise NotebookDifference(
                            "Metadata differ on {} cell #{}: {}\nCell content:\n{}".format(
                                test_cell.cell_type, i, str(error), ref_cell.source
                            )
                        )
                else:
                    modified_cell_metadata.update(set(test_metadata).difference(ref_metadata))
                    modified_cell_metadata.update(set(ref_metadata).difference(test_metadata))
                    for key in set(ref_metadata).intersection(test_metadata):
                        if ref_metadata[key] != test_metadata[key]:
                            modified_cell_metadata.add(key)

        test_lines.extend([line for line in test_cell.source.splitlines() if not _BLANK_LINE.match(line)])

        # 3. test cell content
        if ref_lines != test_lines:
            if raise_on_first_difference:
                try:
                    compare("\n".join(test_lines), "\n".join(ref_lines))
                except AssertionError as error:
                    raise NotebookDifference(f"Cell content differ on {test_cell.cell_type} cell #{i}: {str(error)}")
            else:
                modified_cells.add(i)

        # 3. bis test entire cell content
        if not same_content(ref_cell.source, test_cell.source, allow_removed_final_blank_line):
            if ref_cell.source != test_cell.source:
                if raise_on_first_difference:
                    diff = compare(test_cell.source, ref_cell.source, return_diff=True)
                    raise NotebookDifference(f"Cell content differ on {test_cell.cell_type} cell #{i}: {diff}")
                modified_cells.add(i)

        if not compare_outputs:
            continue

        if ref_cell.cell_type != "code":
            continue

        ref_cell = filtered_cell(
            ref_cell,
            preserve_outputs=compare_outputs,
            cell_metadata_filter=cell_metadata_filter,
        )
        test_cell = filtered_cell(
            test_cell,
            preserve_outputs=compare_outputs,
            cell_metadata_filter=cell_metadata_filter,
        )

        try:
            compare(test_cell, ref_cell)
        except AssertionError as error:
            if raise_on_first_difference:
                raise NotebookDifference(
                    "Cell outputs differ on {} cell #{}: {}".format(test_cell["cell_type"], i, str(error))
                )
            modified_cells.add(i)

    # More cells in the actual notebook?
    remaining_cell_count = 0
    while True:
        try:
            test_cell = next(test_cell_iter)
            if raise_on_first_difference:
                raise NotebookDifference(f"Additional {test_cell.cell_type} cell: {test_cell.source}")
            remaining_cell_count += 1
        except StopIteration:
            break

    if remaining_cell_count and not raise_on_first_difference:
        modified_cells.update(
            range(
                len(expected_cells) + 1,
                len(expected_cells) + 1 + remaining_cell_count,
            )
        )

    return modified_cells, modified_cell_metadata


def test_round_trip_conversion(notebook, fmt, update, allow_expected_differences=True, stop_on_first_error=True):
    """Test round trip conversion for a Jupyter notebook"""
    text = writes(notebook, fmt)
    round_trip = reads(text, fmt)

    if update:
        round_trip = combine_inputs_with_outputs(round_trip, notebook, fmt)

    compare_notebooks(
        round_trip,
        notebook,
        fmt,
        allow_expected_differences,
        raise_on_first_difference=stop_on_first_error,
    )


# The functions below are used in the Jupytext text collection
def create_mirror_file_if_missing(mirror_file, notebook, fmt):
    if not os.path.isfile(mirror_file):
        write(notebook, mirror_file, fmt=fmt)


def assert_conversion_same_as_mirror(nb_file, fmt, mirror_name, compare_notebook=False):
    """This function is used in the tests"""
    dirname, basename = os.path.split(nb_file)
    file_name, org_ext = os.path.splitext(basename)
    fmt = long_form_one_format(fmt)
    notebook = read(nb_file, fmt=fmt)
    fmt = check_auto_ext(fmt, notebook.metadata, "")
    ext = fmt["extension"]
    mirror_file = os.path.join(dirname, "..", "..", "outputs", mirror_name, full_path(file_name, fmt))

    # it's better not to have Jupytext metadata in test notebooks:
    if fmt == "ipynb" and "jupytext" in notebook.metadata:  # pragma: no cover
        notebook.metadata.pop("jupytext")
        write(nb_file, fmt=fmt)

    create_mirror_file_if_missing(mirror_file, notebook, fmt)

    # Compare the text representation of the two notebooks
    if compare_notebook:
        # Read and convert the mirror file to the latest nbformat version if necessary
        nb_mirror = read(mirror_file, as_version=notebook.nbformat)
        nb_mirror.nbformat_minor = notebook.nbformat_minor
        compare_notebooks(nb_mirror, notebook)
        return
    elif ext == ".ipynb":
        notebook = read(mirror_file)
        fmt.update({"extension": org_ext})
        actual = writes(notebook, fmt)
        with open(nb_file, encoding="utf-8") as fp:
            expected = fp.read()
    else:
        actual = writes(notebook, fmt)
        with open(mirror_file, encoding="utf-8") as fp:
            expected = fp.read()

    if not actual.endswith("\n"):
        actual = actual + "\n"

    if fmt.get("format_name") == "marimo":
        lines = expected.splitlines()
        lines = [
            # mirror files were generated with marimo 0.17.8
            f'__generated_with = "{marimo_version()}"' if line == '__generated_with = "0.17.8"' else line
            for line in lines
        ]
        expected = "\n".join(lines) + "\n"

    compare(actual, expected)

    # Compare the two notebooks
    if ext != ".ipynb":
        notebook = read(nb_file)
        nb_mirror = read(mirror_file, fmt=fmt)

        if fmt.get("format_name") == "sphinx":
            nb_mirror.cells = nb_mirror.cells[1:]
            for cell in notebook.cells:
                cell.metadata = {}
            for cell in nb_mirror.cells:
                cell.metadata = {}

        compare_notebooks(nb_mirror, notebook, fmt)

        nb_mirror = combine_inputs_with_outputs(nb_mirror, notebook)
        compare_notebooks(nb_mirror, notebook, fmt, compare_outputs=True)


def notebook_model(nb):
    """Return a notebook model, with content a
    dictionary rather than a notebook object.
    To be used in tests only."""
    return dict(type="notebook", content=json.loads(json.dumps(nb)))



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    This cell uses no particular cell marker

    $$
    \begin{align}
    \dot{x} & = \sigma(y-x)\\
    \dot{y} & = \rho x - y - xz \\
    \dot{z} & = -\beta z + xy
    \end{align}
    $$
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    This cell uses no particular cell marker, and a single slash in the $\LaTeX$ equation

    $$
    \begin{align}
    \dot{x} & = \sigma(y-x) \
    \dot{y} & = \rho x - y - xz \
    \dot{z} & = -\beta z + xy
    \end{align}
    $$
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    This cell uses the triple quote cell markers introduced at https://github.com/mwouts/jupytext/issues/305

    $$
    \begin{align}
    \dot{x} & = \sigma(y-x)\\
    \dot{y} & = \rho x - y - xz \\
    \dot{z} & = -\beta z + xy
    \end{align}
    $$
    """)
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.18.0-dev
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# This paired notebook can be edited either as
# a Python script or as a Jupyter notebook in
# VS Code. Thanks to the Jupytext Sync
# extension, changes are automatically synced to
# the other paired file(s) when you save.
#
# 💡 Tip: don't forget to hit "Save" before
# switching to the other editor. If you don't save
# your changes, you will run into conflicting edits
# as VS Code does not autoreload files that have
# unsaved edits.

# %%
import plotly.express as px

# %%
px.scatter(
    px.data.iris(),
    x="sepal_width",
    y="sepal_length",
    color="species",
)

# %%



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell
def _():
    1 + 1
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    A markdown cell
    And below, the cell for function f has non trivial cell metadata. And the next cell as well.
    """)
    return


@app.function
def f(x):
    return x


@app.cell
def _():
    f(5)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    More text
    """)
    return


@app.cell
def _():
    2 + 2
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()



"""Find and read Jupytext configuration files"""

import json
import os
from pathlib import Path

try:
    import tomllib
except ImportError:
    import tomli as tomllib

import warnings

import yaml
from traitlets import Bool, Dict, Enum, Float, List, Unicode, Union
from traitlets.config import Configurable
from traitlets.config.loader import PyFileConfigLoader
from traitlets.traitlets import TraitError
import typing
from .formats import (
    NOTEBOOK_EXTENSIONS,
    get_formats_from_notebook_metadata,
    long_form_multiple_formats,
    long_form_one_format,
    short_form_multiple_formats,
)


class JupytextConfigurationError(ValueError):
    """Error in the specification of the format for the text notebook"""


JUPYTEXT_CONFIG_FILES = [
    "jupytext",
    "jupytext.toml",
    "jupytext.yml",
    "jupytext.yaml",
    "jupytext.json",
]

JUPYTEXT_CONFIG_FILES.extend(["." + filename for filename in JUPYTEXT_CONFIG_FILES] + [".jupytext.py"])

PYPROJECT_FILE = "pyproject.toml"

JUPYTEXT_CEILING_DIRECTORIES = [path for path in os.environ.get("JUPYTEXT_CEILING_DIRECTORIES", "").split(":") if path]


class JupytextConfiguration(Configurable):
    """Jupytext Configuration's options"""

    formats = Union(
        [Unicode(), List(Unicode()), List(Dict(Unicode())), Dict(Unicode())],
        help="Save notebooks to these file extensions. "
        "Can be any of ipynb,Rmd,md,jl,py,R,nb.jl,nb.py,nb.R "
        "comma separated. If you want another format than the "
        "default one, append the format name to the extension, "
        "e.g. ipynb,py:percent to save the notebook to "
        "hydrogen/spyder/vscode compatible scripts. "
        "Can also be a list of format dictionaries for first-match pairing.",
        config=True,
    )
    default_jupytext_formats = Unicode(help="Deprecated. Use 'formats' instead", config=True)

    preferred_jupytext_formats_save = Unicode(
        help="Preferred format when saving notebooks as text, per extension. "
        'Use "jl:percent,py:percent,R:percent" if you want to save '
        "Julia, Python and R scripts in the double percent format and "
        'only write "jupytext_formats": "py" in the notebook metadata.',
        config=True,
    )

    preferred_jupytext_formats_read = Unicode(
        help="Preferred format when reading notebooks from text, per "
        'extension. Use "py:sphinx" if you want to read all python '
        "scripts as Sphinx gallery scripts.",
        config=True,
    )

    notebook_metadata_filter = Unicode(
        help="Notebook metadata that should be save in the text representations. "
        "Examples: 'all', '-all', 'widgets,nteract', 'kernelspec,jupytext-all'",
        config=True,
    )

    default_notebook_metadata_filter = Unicode("", help="Deprecated. Use 'notebook_metadata_filter' instead", config=True)

    hide_notebook_metadata = Enum(
        values=[True, False],
        allow_none=True,
        help="Should the notebook metadata be wrapped into an HTML comment in the Markdown format?",
        config=True,
    )

    root_level_metadata_as_raw_cell = Bool(
        True,
        help="Should the root level metadata of text documents (like the fields 'title' or 'author' in "
        "R Markdown document) appear as a raw cell in the notebook (True), or go to the notebook"
        "metadata?",
        config=True,
    )

    root_level_metadata_filter = Unicode(
        help="Notebook metadata that should be promoted to the root level in the text representations. "
        "Examples: 'all', '-all', 'kernelspec,jupytext'",
        config=True,
    )

    cell_metadata_filter = Unicode(
        help="Cell metadata that should be saved in the text representations. Examples: 'all', 'hide_input,hide_output'",
        config=True,
    )

    default_cell_metadata_filter = Unicode("", help="Deprecated. Use 'cell_metadata_filter' instead", config=True)

    comment_magics = Enum(
        values=[True, False],
        allow_none=True,
        help="Should Jupyter magic commands be commented out in the text representation?",
        config=True,
    )

    split_at_heading = Bool(
        False,
        help="Split markdown cells on headings (Markdown and R Markdown formats only)",
        config=True,
    )

    sphinx_convert_rst2md = Bool(
        False,
        help="When opening a Sphinx Gallery script, convert the reStructuredText to markdown",
        config=True,
    )

    doxygen_equation_markers = Bool(
        False,
        help="Should equation markers use the DOxygen format? (see https://github.com/mwouts/jupytext/issues/517)",
        config=True,
    )

    outdated_text_notebook_margin = Float(
        1.0,
        help="Refuse to overwrite inputs of a ipynb notebooks with those of a "
        "text notebook when the text notebook plus margin is older than "
        "the ipynb notebook (NB: This option is ignored by Jupytext CLI)",
        config=True,
    )

    cm_config_log_level = Enum(
        values=["warning", "info", "info_if_changed", "debug", "none"],
        default_value="info_if_changed",
        help="The log level for config file logs in the Jupytext contents manager",
        config=True,
    )

    cell_markers = Unicode(
        help='Start and end cell markers for the light format, comma separated. Use "{{{,}}}" to mark cells'
        'as foldable regions in Vim, and "region,endregion" to mark cells as Vscode/PyCharm regions',
        config=True,
    )

    default_cell_markers = Unicode(help="Deprecated. Use 'cell_markers' instead", config=True)

    notebook_extensions = Union(
        [List(Unicode(), NOTEBOOK_EXTENSIONS), Unicode()],
        help="A list of notebook extensions",
        config=True,
    )

    custom_cell_magics = Unicode(
        help='A comma separated list of cell magics. Use e.g. custom_cell_magics = "configure,local" '
        'if you want code cells starting with the Spark magic cell commands "configure" and "local" '
        "to be commented out when converted to scripts.",
        config=True,
    )

    def set_default_format_options(self, format_options, read=False):
        """Set default format option"""
        if self.default_notebook_metadata_filter:
            warnings.warn(
                "The option 'default_notebook_metadata_filter' is deprecated. Please use 'notebook_metadata_filter' instead.",
                FutureWarning,
            )
            format_options.setdefault("notebook_metadata_filter", self.default_notebook_metadata_filter)
        if self.notebook_metadata_filter:
            format_options.setdefault("notebook_metadata_filter", self.notebook_metadata_filter)
        if self.default_cell_metadata_filter:
            warnings.warn(
                "The option 'default_cell_metadata_filter' is deprecated. Please use 'cell_metadata_filter' instead.",
                FutureWarning,
            )
            format_options.setdefault("cell_metadata_filter", self.default_cell_metadata_filter)
        if self.root_level_metadata_filter:
            format_options.setdefault("root_level_metadata_filter", self.root_level_metadata_filter)
        if self.cell_metadata_filter:
            format_options.setdefault("cell_metadata_filter", self.cell_metadata_filter)
        if self.hide_notebook_metadata is not None:
            format_options.setdefault("hide_notebook_metadata", self.hide_notebook_metadata)
        if self.root_level_metadata_as_raw_cell is False:
            format_options.setdefault("root_level_metadata_as_raw_cell", self.root_level_metadata_as_raw_cell)
        if self.comment_magics is not None:
            format_options.setdefault("comment_magics", self.comment_magics)
        if self.split_at_heading:
            format_options.setdefault("split_at_heading", self.split_at_heading)
        if self.doxygen_equation_markers:
            format_options.setdefault("doxygen_equation_markers", self.doxygen_equation_markers)
        if not read:
            if self.default_cell_markers:
                warnings.warn(
                    "The option 'default_cell_markers' is deprecated. Please use 'cell_markers' instead.",
                    FutureWarning,
                )
                format_options.setdefault("cell_markers", self.default_cell_markers)
            if self.cell_markers:
                format_options.setdefault("cell_markers", self.cell_markers)
        if read and self.sphinx_convert_rst2md:
            format_options.setdefault("rst2md", self.sphinx_convert_rst2md)
        if self.custom_cell_magics:
            format_options.setdefault("custom_cell_magics", self.custom_cell_magics)

    def default_formats(self, path):
        """Return the default formats, if they apply to the current path #157"""
        from .paired_paths import InconsistentPath, base_path

        if self.default_jupytext_formats:
            warnings.warn(
                "The option 'default_jupytext_formats' is deprecated. Please use 'formats' instead.",
                FutureWarning,
            )

        # formats is a list of paired formats - find the first match
        for paired_formats in normalize_formats(self.formats or self.default_jupytext_formats):
            # Check if one of the paired format matches the current path
            for fmt in long_form_multiple_formats(paired_formats):
                try:
                    base_path(path, fmt)
                    return paired_formats
                except InconsistentPath:
                    continue
        return None

    def __eq__(self, other):
        for key in self.class_trait_names():
            if getattr(self, key) != getattr(other, key):
                return False

        return True


def preferred_format(incomplete_format, preferred_formats):
    """Return the preferred format for the given extension"""
    incomplete_format = long_form_one_format(incomplete_format)
    if "format_name" in incomplete_format:
        return incomplete_format

    for fmt in long_form_multiple_formats(preferred_formats):
        if (
            (
                incomplete_format["extension"] == fmt["extension"]
                or (
                    fmt["extension"] == ".auto"
                    and incomplete_format["extension"] not in [".md", ".markdown", ".Rmd", ".ipynb"]
                )
            )
            and incomplete_format.get("suffix") == fmt.get("suffix", incomplete_format.get("suffix"))
            and incomplete_format.get("prefix") == fmt.get("prefix", incomplete_format.get("prefix"))
        ):
            fmt.update(incomplete_format)
            return fmt

    return incomplete_format


def global_jupytext_configuration_directories():
    """Return the directories in which Jupytext will search for a configuration file"""

    config_dirs = []

    if "XDG_CONFIG_HOME" in os.environ:
        config_dirs.extend(os.environ["XDG_CONFIG_HOME"].split(":"))
    elif "USERPROFILE" in os.environ:
        config_dirs.append(os.environ["USERPROFILE"])
    elif "HOME" in os.environ:
        config_dirs.append(os.path.join(os.environ["HOME"], ".config"))
        config_dirs.append(os.environ["HOME"])

    if "XDG_CONFIG_DIRS" in os.environ:
        config_dirs.extend(os.environ["XDG_CONFIG_DIRS"].split(":"))
    elif "ALLUSERSPROFILE" in os.environ:
        config_dirs.append(os.environ["ALLUSERSPROFILE"])
    else:
        config_dirs.extend(["/usr/local/share/", "/usr/share/"])

    for config_dir in config_dirs:
        yield from [
            os.path.join(config_dir, "jupytext"),
            config_dir,
        ]


def find_global_jupytext_configuration_file():
    """Return the global Jupytext configuration file, if any"""

    for config_dir in global_jupytext_configuration_directories():
        config_file = find_jupytext_configuration_file(config_dir, False)
        if config_file:
            return config_file

    return None


def find_jupytext_configuration_file(path: typing.Union[str, Path], search_parent_dirs=True) -> str:
    """Return the first jupytext configuration file in the current directory, or any parent directory"""

    path = Path(path).absolute()

    if path.is_dir():
        for filename in JUPYTEXT_CONFIG_FILES:
            full_path = path / filename
            if full_path.is_file():
                return str(full_path)

    pyproject_path = path / PYPROJECT_FILE
    if pyproject_path.is_file():
        with pyproject_path.open() as stream:
            doc = tomllib.loads(stream.read())
            if doc.get("tool", {}).get("jupytext") is not None:
                return str(pyproject_path)

    if not search_parent_dirs:
        return None

    if JUPYTEXT_CEILING_DIRECTORIES and path.is_dir():
        for ceiling_dir in JUPYTEXT_CEILING_DIRECTORIES:
            if Path(ceiling_dir).is_dir() and path.absolute() == Path(ceiling_dir).absolute():
                return None

    parent_dir = path.parent
    if parent_dir == path:
        return find_global_jupytext_configuration_file()

    return find_jupytext_configuration_file(parent_dir, True)


def parse_jupytext_configuration_file(jupytext_config_file, stream=None):
    """Read a Jupytext config file, and return a dict"""
    if not jupytext_config_file.endswith(".py") and stream is None:
        with open(jupytext_config_file, encoding="utf-8") as stream:
            return parse_jupytext_configuration_file(jupytext_config_file, stream.read())

    try:
        if jupytext_config_file.endswith((".toml", "jupytext")):
            doc = tomllib.loads(stream)
            if jupytext_config_file.endswith(PYPROJECT_FILE):
                return doc["tool"]["jupytext"]
            else:
                return doc

        if jupytext_config_file.endswith((".yml", ".yaml")):
            return yaml.safe_load(stream)

        if jupytext_config_file.endswith(".json"):
            return json.loads(stream)

        return PyFileConfigLoader(jupytext_config_file).load_config()
    except (ValueError, NameError) as err:
        raise JupytextConfigurationError(f"The Jupytext configuration file {jupytext_config_file} is incorrect: {err}")


def normalize_formats(formats) -> list[str]:
    """Normalize the formats option into a list of string-encoded paired formats"""
    # Process formats - can be string, dict, or list
    if isinstance(formats, str):
        # Split on semicolon for multiple format groups
        formats = formats.split(";")
    elif isinstance(formats, dict):
        # Single dict - wrap in list for uniform processing
        formats = [formats]
    elif formats is None:
        formats = []
    elif not isinstance(formats, list):
        raise JupytextConfigurationError(
            f"Invalid type for 'formats': {type(formats).__name__}. Expected str, dict, list of str or dict."
        )

    # Each group of paired formats can be a string or a dict
    string_encoded_pairing_formats = []

    for paired_formats in formats:
        if isinstance(paired_formats, str):
            string_encoded_pairing_formats.append(paired_formats)
        elif isinstance(paired_formats, dict):
            # Convert dict to format string
            paired_formats = [
                (f if not prefix else (prefix[:-1] if prefix.endswith("/") else prefix) + "///" + f)
                for prefix, f in paired_formats.items()
            ]
            string_encoded_pairing_formats.append(short_form_multiple_formats(paired_formats))
        else:
            raise JupytextConfigurationError(
                f"Invalid paired formats: {paired_formats}. Expected str or dict, got {type(paired_formats).__name__}."
            )

    return string_encoded_pairing_formats


def load_jupytext_configuration_file(config_file, stream=None):
    """Read and validate a Jupytext configuration file, and return a JupytextConfiguration object"""
    config_dict = parse_jupytext_configuration_file(config_file, stream)
    config = validate_jupytext_configuration_file(config_file, config_dict)
    config.formats = normalize_formats(config.formats or config.default_jupytext_formats)

    if isinstance(config.notebook_extensions, str):
        config.notebook_extensions = config.notebook_extensions.split(",")
    return config


def load_jupytext_config(nb_file):
    """Return the jupytext configuration file in the same folder, or in a parent folder, of the current file, if any"""
    config_file = find_jupytext_configuration_file(nb_file)
    if config_file is None:
        return None
    if os.path.isfile(nb_file) and os.path.samefile(config_file, nb_file):
        return None
    config_file = find_jupytext_configuration_file(nb_file)
    return load_jupytext_configuration_file(config_file)


def validate_jupytext_configuration_file(config_file, config_dict):
    """Turn a dict-like config into a JupytextConfiguration object"""
    if config_dict is None:
        return None
    try:
        config = JupytextConfiguration(**config_dict)
    except TraitError as err:
        raise JupytextConfigurationError(f"The Jupytext configuration file {config_file} is incorrect: {err}")
    invalid_options = set(config_dict).difference(dir(JupytextConfiguration()))
    if any(invalid_options):
        raise JupytextConfigurationError(
            "The Jupytext configuration file {} is incorrect: options {} are not supported".format(
                config_file, ",".join(invalid_options)
            )
        )
    return config


def notebook_formats(nbk, config, path, fallback_on_current_fmt=True):
    """Return the list of formats for the current notebook"""
    metadata = nbk.get("metadata")
    jupytext_metadata = metadata.get("jupytext", {})
    formats = jupytext_metadata.get("formats") or metadata.get("jupytext_formats")

    if formats:
        formats = long_form_multiple_formats(formats, metadata, auto_ext_requires_language_info=False)
    elif config:
        current_format = jupytext_metadata.get("text_representation", {"extension": os.path.splitext(path)[1]})
        default_formats = long_form_multiple_formats(
            config.default_formats(path),
            metadata,
            auto_ext_requires_language_info=False,
        )

        if any(
            current_format.get("extension") == fmt["extension"]
            and (
                "format_name" not in fmt
                or "format_name" not in current_format
                or current_format["format_name"] == fmt.get("format_name")
            )
            for fmt in default_formats
        ):
            formats = default_formats

    if not formats:
        if not fallback_on_current_fmt:
            return None
        text_repr = jupytext_metadata.get("text_representation", {})
        ext = os.path.splitext(path)[1]
        fmt = {"extension": ext}

        if ext == text_repr.get("extension") and text_repr.get("format_name"):
            fmt["format_name"] = text_repr.get("format_name")

        formats = [fmt]

    # Set preferred formats if no format name has been given yet
    if config:
        formats = [preferred_format(f, config.preferred_jupytext_formats_save) for f in formats]

    return formats


def get_formats_from_notebook_and_config(notebook, config, nb_file):
    """
    Get the notebook formats from notebook metadata or config.

    Notebook metadata takes precedence over config. If the notebook metadata contains pairing information,
    it is used; otherwise, the configuration is used as a fallback.

    Parameters
    ----------
    notebook : dict
        The notebook object (as a dictionary).
    config : JupytextConfiguration or None
        The Jupytext configuration object.
    nb_file : str
        The path to the notebook file.

    Returns
    -------
    list
        A list of format dictionaries describing the notebook's paired formats.
    """
    formats = get_formats_from_notebook_metadata(notebook)
    if formats:
        return long_form_multiple_formats(formats)
    else:
        return notebook_formats(notebook, config, nb_file)



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell
def _():
    # This is an unfrozen cell. Works as usual.
    print("I'm a regular cell so I run and print!")
    return


@app.cell
def _():
    # This is an frozen cell
    print("I'm frozen so Im not executed :(")
    return


if __name__ == "__main__":
    app.run()



"""
Here we test the --check-source-is-newer option of the jupytext CLI
"""

import os
from jupytext import cli
from jupytext import write

import pytest


def test_check_source_is_newer_when_using_jupytext_to(tmp_path, python_notebook):
    tmp_ipynb = tmp_path / "notebook.ipynb"
    tmp_py = tmp_path / "notebook.py"
    write(python_notebook, tmp_ipynb, fmt="ipynb")

    # First, we convert the .ipynb to .py
    cli.jupytext([str(tmp_ipynb), "--to", "py", "--check-source-is-newer"])
    assert tmp_py.exists()

    # Now, we modify the .py file so that it is more recent than the .ipynb file
    text = tmp_py.read_text() + "\n# A new comment"
    tmp_py.write_text(text)

    # We can convert the .py to .ipynb because the .py is more recent than the .ipynb
    cli.jupytext([str(tmp_py), "--to", "ipynb", "--check-source-is-newer"])

    # We modify the .ipynb file so that it is more recent than the .py file
    text = tmp_ipynb.read_text().replace("A new comment", "Another new comment")
    tmp_ipynb.write_text(text)

    # Now, trying to convert the .py to .ipynb raises an error because the .py is older than the .ipynb
    with pytest.raises(ValueError, match=r"Source .*notebook\.py.* is older than destination .*notebook\.ipynb.*"):
        cli.jupytext([str(tmp_py), "--to", "ipynb", "--check-source-is-newer"])


def test_check_source_is_newer_when_using_jupytext_sync(tmp_path, python_notebook):
    tmp_ipynb = tmp_path / "notebook.ipynb"
    tmp_py = tmp_path / "notebook.py"
    write(python_notebook, tmp_ipynb, fmt="ipynb")

    # First, we turn the notebook into a paired notebook
    cli.jupytext([str(tmp_ipynb), "--set-formats", "ipynb,py", "--check-source-is-newer"])
    assert tmp_py.exists()

    # Running sync on the .py file works as .py is always more recent after a --sync operation
    cli.jupytext([str(tmp_py), "--sync", "--check-source-is-newer"])

    # Make .ipynb slightly older again to ensure .py is newer
    stat = os.stat(tmp_ipynb)
    os.utime(tmp_ipynb, (stat.st_atime, stat.st_mtime - 1))

    # Now, trying to sync the .ipynb to .py raises an error because .ipynb is older than .py
    with pytest.raises(ValueError, match=r"Source .*notebook\.ipynb.* is older than paired file .*notebook\.py.*"):
        cli.jupytext([str(tmp_ipynb), "--sync", "--check-source-is-newer"])

    # We modify the .ipynb file so that it is more recent than the .py file
    text = tmp_ipynb.read_text().replace("A short notebook", "A short notebook with a modification")
    tmp_ipynb.write_text(text)

    # Make .py slightly older to ensure .ipynb is newer
    stat = os.stat(tmp_py)
    os.utime(tmp_py, (stat.st_atime, stat.st_mtime - 1))

    # Now, trying to sync the .py to .ipynb raises an error because the .py is older than the .ipynb
    with pytest.raises(ValueError, match=r"Source .*notebook\.py.* is older than paired file .*notebook\.ipynb.*"):
        cli.jupytext([str(tmp_py), "--sync", "--check-source-is-newer"])

    # Running sync on the .ipynb file works as .ipynb is now more recent than .py
    cli.jupytext([str(tmp_ipynb), "--sync", "--check-source-is-newer"])



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Jupyter notebook

    This notebook is a simple jupyter notebook. It only has markdown and code cells. And it does not contain consecutive markdown cells. We start with an addition:
    """)
    return


@app.cell
def _():
    a = 1
    b = 2
    a + b
    return a, b


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    Now we return a few tuples
    """)
    return


@app.cell
def _(a, b):
    a, b
    return


@app.cell
def _(a, b):
    a, b, a+b
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    And this is already the end of the notebook
    """)
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()



import pytest
from git.exc import HookExecutionError
from nbformat.v4.nbbase import new_code_cell, new_markdown_cell, new_notebook
from pre_commit.main import main as pre_commit

from jupytext import read, write


@pytest.mark.requires_pandoc
def test_pre_commit_hook_sync_reformat_code_and_markdown(
    tmpdir,
    cwd_tmpdir,
    tmp_repo,
    jupytext_repo_root,
    jupytext_repo_rev,
    notebook_with_outputs,
):
    """Here we sync the ipynb notebook with a py:percent file and also apply black and pandoc to reformat both
    code and markdown cells.
    """
    pre_commit_config_yaml = f"""
repos:
- repo: {jupytext_repo_root}
  rev: {jupytext_repo_rev}
  hooks:
  - id: jupytext
    args: [--sync, --pipe-fmt, ipynb, --pipe, 'pandoc --from ipynb --to ipynb --markdown-headings=atx', --show-changes]
  - id: jupytext
    args: [--sync, --pipe, black, --show-changes]
    additional_dependencies:
    - black==22.3.0  # Matches black hook below

- repo: https://github.com/psf/black
  rev: 22.3.0
  hooks:
  - id: black
"""
    tmpdir.join(".pre-commit-config.yaml").write(pre_commit_config_yaml)

    tmp_repo.git.add(".pre-commit-config.yaml")
    pre_commit(["install", "--install-hooks", "-f"])

    tmpdir.join("jupytext.toml").write('formats = "ipynb,py:percent"')
    tmp_repo.git.add("jupytext.toml")
    tmp_repo.index.commit("pair notebooks")

    # write a test notebook
    notebook = new_notebook(
        cells=[
            new_code_cell("1+1"),
            new_markdown_cell(
                """This is a complex markdown cell

# With a h1 header
## And a h2 header

| And       | A  | Table |
| --------- | ---| ----- |
| 0         | 1  | 2     |

!!!WARNING!!! This hook does not seem compatible with
explicit paragraph breaks (two spaces at the end of a line).

And a VERY long line.
""".replace("VERY ", "very " * 51)
            ),
        ],
        metadata=notebook_with_outputs.metadata,
    )

    write(notebook, "test.ipynb")

    # try to commit it, should fail because
    # 1. the py version hasn't been added
    # 2. the first cell is '1+1' which is not black compliant
    # 3. the second cell needs to be wrapped
    tmp_repo.git.add("test.ipynb")
    with pytest.raises(
        HookExecutionError,
        match="files were modified by this hook",
    ):
        tmp_repo.index.commit("failing")

    # Add the two files
    tmp_repo.git.add("test.ipynb")
    tmp_repo.git.add("test.py")

    # now the commit will succeed
    tmp_repo.index.commit("passing")
    assert "test.ipynb" in tmp_repo.tree()
    assert "test.py" in tmp_repo.tree()

    # both the code and the markdown cells were reformatted
    nb = read("test.ipynb")
    assert nb.cells[0].source == "1 + 1"

    print(nb.cells[1].source)
    assert (
        nb.cells[1].source
        == """This is a complex markdown cell

# With a h1 header

## And a h2 header

| And | A   | Table |
|-----|-----|-------|
| 0   | 1   | 2     |

!!!WARNING!!! This hook does not seem compatible with explicit paragraph
breaks (two spaces at the end of a line).

And a very very very very very very very very very very very very very
very very very very very very very very very very very very very very
very very very very very very very very very very very very very very
very very very very very very very very very very long line."""
    )



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    This is a simple marimo notebook
    """)
    return


@app.cell
def _():
    x = 1
    return (x,)


@app.cell
def _(x):
    y = x+1
    y
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()



"""Jupyter notebook to Marimo py format and back, using Marimo"""

import os
from packaging.version import parse
import tempfile
import subprocess
import nbformat

MARIMO_MIN_VERSION = "0.16.3"


class MarimoError(OSError):
    """An error related to Marimo"""


def is_marimo_available(min_version=MARIMO_MIN_VERSION, max_version=None):
    """Is Marimo available?"""
    try:
        raise_if_marimo_is_not_available(min_version=min_version, max_version=max_version)
        return True
    except MarimoError:
        return False


def raise_if_marimo_is_not_available(min_version=MARIMO_MIN_VERSION, max_version=None):
    """Raise with an informative error message if Marimo is not available"""
    version = marimo_version()
    if version == "N/A":
        raise MarimoError(f"The Marimo format requires 'marimo>={min_version}', but marimo was not found")

    if parse(version) < parse(min_version):
        raise MarimoError(f"The Marimo format requires 'marimo>={min_version}', but marimo=={version} was found")

    if max_version and parse(version) > parse(max_version):
        raise MarimoError(f"The Marimo format requires 'marimo<={max_version}', but marimo=={version} was found")

    return version


def marimo_version():
    """Marimo's version number"""
    try:
        return marimo("--version").strip()
    except OSError:
        return "N/A"


def marimo_py_to_notebook(text):
    """Convert a Marimo script to a Jupyter notebook, using Marimo"""
    raise_if_marimo_is_not_available()

    # On Windows, NamedTemporaryFile cannot be reopened with open,
    # so we keep the file names and close the files
    tmp_py_file = tempfile.NamedTemporaryFile(suffix=".py")
    tmp_py_file_name = tmp_py_file.name
    tmp_py_file.close()

    tmp_ipynb_file = tempfile.NamedTemporaryFile(suffix=".ipynb")
    tmp_ipynb_file_name = tmp_ipynb_file.name
    tmp_ipynb_file.close()

    with open(tmp_py_file_name, "w") as fp:
        fp.write(text)

    marimo(
        "export",
        "ipynb",
        # Keep the current order to minimize diffs on round trips
        "--sort",
        "top-down",
        tmp_py_file_name,
        "-o",
        tmp_ipynb_file_name,
    )

    notebook = nbformat.read(tmp_ipynb_file_name, as_version=4)

    os.remove(tmp_py_file_name)
    os.remove(tmp_ipynb_file_name)

    # In the following we revert some of the side effects of the marimo conversion
    # to ensure stability of the round trip. Not all the side effects are reverted.

    # You can test the round trip for a given document with:
    #   jupytext --test --to py:marimo your_notebook.ipynb
    # or with:
    #   jupytext --test --to ipynb your_marimo_script.py

    # Ideally these round trip issues should be fixed in Marimo itself - please report them at
    # https://github.com/marimo/marimo/issues and optionally mention @mwouts (author of Jupytext)
    # in the issue description.
    import_marimo_cell = "import marimo as mo"
    need_to_remove_import_marimo_cell = False

    for cell in notebook.cells:
        if cell.cell_type != "code":
            continue
        matplotlib_inline_comment = "# '%matplotlib inline' command supported automatically in marimo"
        if cell.source.startswith(matplotlib_inline_comment):
            cell.source = "%matplotlib inline" + cell.source.removeprefix(matplotlib_inline_comment)
        if cell.source == import_marimo_cell:
            need_to_remove_import_marimo_cell = True
        if cell.source.startswith("# Cell tags:"):
            if "\n" not in cell.source:
                cell.source += "\n"
            tag_line, cell.source = cell.source.split("\n", 1)
            tags = tag_line.removeprefix("# Cell tags:").strip().split(", ")
            cell.metadata["tags"] = tags

    if need_to_remove_import_marimo_cell:
        notebook.cells = [cell for cell in notebook.cells if cell.source != import_marimo_cell]

    return notebook


def notebook_to_marimo_py(notebook):
    """Convert a notebook to its Marimo script"""
    raise_if_marimo_is_not_available()

    # On Windows, NamedTemporaryFile cannot be reopened with open,
    # so we keep the file names and close the files
    tmp_py_file = tempfile.NamedTemporaryFile(suffix=".py", delete=False)
    tmp_py_file_name = tmp_py_file.name
    tmp_py_file.close()

    tmp_ipynb_file = tempfile.NamedTemporaryFile(suffix=".ipynb", delete=False)
    tmp_ipynb_file_name = tmp_ipynb_file.name
    tmp_ipynb_file.close()

    with open(tmp_ipynb_file_name, "w") as fp:
        nbformat.write(notebook, fp)

    marimo("convert", tmp_ipynb_file_name, "-o", tmp_py_file_name)

    with open(tmp_py_file_name) as fp:
        text = fp.read()

    os.remove(tmp_ipynb_file_name)
    os.remove(tmp_py_file_name)

    return "\n".join(text.splitlines())


def marimo(*args: str):
    cmd = ["marimo"] + list(args)

    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    out, err = proc.communicate()
    if proc.returncode:
        raise MarimoError(f"marimo exited with return code {proc.returncode}\n{str(err)}")

    return out.decode("utf-8")



"""
These tests ensure that Jupytext raises an error when a file loaded by Jupytext
changes while Jupytext is running.

To make the simultaneous change occur in the tests, we monkey-patch the function `create_prefix_dir`
which is called just before writing the file back to disk.
"""

from jupytext import cli
from jupytext import write

import pytest


def test_jupytext_sync_raises_on_synchronous_edits(tmp_path, python_notebook, monkeypatch):
    (tmp_path / "jupytext.toml").write_text("""formats = "ipynb,py" """)

    tmp_py = tmp_path / "notebook.py"
    tmp_ipynb = tmp_path / "notebook.ipynb"
    write(python_notebook, tmp_py, fmt="py:percent")

    def edit_py_file(path):
        text = tmp_py.read_text() + "\n# A new comment"
        tmp_py.write_text(text)

    # The print statements help to see which Jupytext actions correspond to which test
    print("Testing a synchronous edit on a .py file that does not have a .ipynb counterpart")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", edit_py_file)
    with pytest.raises(
        cli.SynchronousModificationError, match=r"The file .*notebook\.py.* was modified while Jupytext was running"
    ):
        cli.jupytext([str(tmp_py), "--sync"])

    assert not tmp_ipynb.exists()

    print("Testing the normal sync command that creates the .ipynb file")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", None)
    cli.jupytext([str(tmp_py), "--sync"])
    assert tmp_ipynb.exists()

    print("Testing a synchronous edit on a .py file that does have a .ipynb counterpart")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", edit_py_file)
    with pytest.raises(
        cli.SynchronousModificationError, match=r"The file .*notebook\.py.* was modified while Jupytext was running"
    ):
        cli.jupytext([str(tmp_py), "--sync"])

    print("Testing a synchronous deletion of the .ipynb file")

    def rm_ipynb_file(path):
        tmp_ipynb.unlink()

    monkeypatch.setattr(cli, "_callback_on_lazy_write", rm_ipynb_file)
    with pytest.raises(
        cli.SynchronousModificationError, match=r"The file .*notebook\.ipynb.* was deleted while Jupytext was running"
    ):
        cli.jupytext([str(tmp_py), "--sync"])

    assert not tmp_ipynb.exists()

    print("Testing a synchronous creation of the .ipynb file")

    def create_ipynb_file(path):
        tmp_ipynb.write_text("{}")

    monkeypatch.setattr(cli, "_callback_on_lazy_write", create_ipynb_file)
    with pytest.raises(
        cli.SynchronousModificationError, match=r"The file .*notebook\.ipynb.* was created while Jupytext was running"
    ):
        cli.jupytext([str(tmp_py), "--sync"])


def test_jupytext_to_raises_on_synchronous_edits(tmp_path, python_notebook, monkeypatch):
    tmp_py = tmp_path / "notebook.py"
    tmp_ipynb = tmp_path / "notebook.ipynb"
    write(python_notebook, tmp_py, fmt="py:percent")

    def edit_py_file(path):
        text = tmp_py.read_text() + "\n# A new comment"
        tmp_py.write_text(text)

    # The print statements help to see which Jupytext actions correspond to which test
    print("Testing a synchronous edit on a .py file")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", edit_py_file)
    with pytest.raises(
        cli.SynchronousModificationError, match=r"The file .*notebook\.py.* was modified while Jupytext was running"
    ):
        cli.jupytext([str(tmp_py), "--to", "ipynb"])

    assert not tmp_ipynb.exists()

    print("Testing the normal to command that creates the .ipynb file")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", None)
    cli.jupytext([str(tmp_py), "--to", "ipynb"])
    assert tmp_ipynb.exists()

    print("Testing a synchronous edit on a .py file that is not paired to a .ipynb file")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", edit_py_file)
    # This one works as the .py file is not paired, so it is not read
    cli.jupytext([str(tmp_ipynb), "--to", "py"])

    print("Turning the .py file into a paired notebook")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", None)
    cli.jupytext([str(tmp_ipynb), "--set-formats", "ipynb,py"])

    print("Testing a synchronous edit on a paired .py file")
    monkeypatch.setattr(cli, "_callback_on_lazy_write", edit_py_file)
    with pytest.raises(
        cli.SynchronousModificationError, match=r"The file .*notebook\.py.* was modified while Jupytext was running"
    ):
        cli.jupytext([str(tmp_py), "--to", "ipynb"])



# ---
# jupyter:
#   jupytext:
#     cell_markers: '{{{,}}}'
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# {{{ [markdown] @deathbeds/jupyterlab-fonts={"styles": {"": {"body[data-jp-deck-mode='presenting'] &": {"right": "0", "top": "30%", "width": "25%", "z-index": 1}}}} jupyterlab-slideshow={"layer": "slide"}
# > **Note**
# > 
# > `slide` layer with a `top` of `30%`
# }}}



import marimo

__generated_with = "0.17.8"
app = marimo.App()


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    This notebook contains complex outputs, including plotly javascript graphs.
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    # Interactive plots
    """)
    return


@app.cell(hide_code=True)
def _(mo):
    mo.md(r"""
    We use Plotly's connected mode to make the notebook lighter - when connected, the notebook downloads the `plotly.js` library from the web.
    """)
    return


@app.cell
def _():
    import plotly.offline as offline
    offline.init_notebook_mode(connected=True)
    return


@app.cell
def _():
    import plotly.graph_objects as go
    fig = go.Figure(
        data=[go.Bar(y=[2, 3, 1])],
        layout=go.Layout(title="bar plot"))
    fig.show()
    fig.data[0].marker = dict(color='purple')
    fig
    return


@app.cell
def _():
    import marimo as mo
    return (mo,)


if __name__ == "__main__":
    app.run()
